local ffi = assert( require 'ffi' )
local buffer = assert( require( "string.buffer" ) )
local ipString = assert( require 'shared.ipstring' )
local roles = assert( require 'shared.roles' )
local packet = {}
local mt = { __index = { new = function( self ) return ffi.new( self.ct ) end } }

local headerByte = 0x41 --Ensure printable characters at start of packets.

local function newStruct( t )
  assert( not( packet[ t.name ] ))
  packet[ t.name ] = t
  t.netname = headerByte
  packet[ headerByte ] = t
  headerByte = headerByte + 1
  ffi.cdef(("typedef struct {uint8_t netname;\n%s;\n} %s;"):format( table.concat( t, ";\n" ), t.name ))
  t.ct = ffi.typeof( ffi.new( t.name ) )
  t.size = ffi.sizeof( t.ct )
  assert( t.size < 500, t.name )
  setmetatable( t, mt )

  print( "Packet:", t.name, "Members:", #t + 1, "Size:", t.size, "Alignment:", ffi.alignof( t.ct ) )
end

newStruct{ 
  name = "serverInfo",
  "uint8_t   players",
  "uint8_t  capacity",
  "ipAddress      ip",
  "uint16_t  version",
  "uint16_t     port",
  "char   svname[32]",
  "char      map[32]",
}

newStruct{
  name = "clientInfo",
  "char username[31]",
}

newStruct{
  name =    "svChimo",
  "uint32_t    nonce",
}

newStruct{
  name =    "clChimo",
  "uint32_t    nonce",
  "uint32_t     hash",
}

newStruct{
  name = "connected",
   "uint32_t  token",
   "uint32_t  tick",
}

newStruct{
  name = "metaServer",
  "char   padding[300]" --Just a bunch of padding to mitigate amplification. 
}

newStruct{
  name =  "insect",
  "uint8_t     id",
  "bool      dead",
  "int16_t      z",
  "int8_t      hp",
  "int8_t      vx",
  "int8_t      vy",
  "int8_t      vz",
  "int32_t      x",
  "int32_t      y",
}

newStruct{
  name = "soleil",
  "uint16_t  azimuth",
  "uint16_t altitude",
  "int8_t       vazi",
  "int8_t       valt",
}

newStruct{
  name =   "playerInfo",
  "uint8_t          id",
  "role_t         role",
  "char   username[31]",
}

newStruct{
  name = "chatMessage",
  "uint8_t         id",
  "char     cmsg[127]",
}

newStruct{
  name = "command",
  "char   command",
}

newStruct{
  name = "advertised",
  "uint32_t     time",
  "ipAddress      ip",
  "uint16_t     port",
}

newStruct{
  name = "disconnect",
  "char       reason",
}

newStruct{
  name = "debugLatency",
  "uint32_t lastTick",
}

local readBuffer = buffer.new( 1024 )
function packet.deserialise( str )
  readBuffer:set( str )
  local data = {}
  local types = {}
  local n = 0
  while #readBuffer ~= 0 do
    local netname = readBuffer:ref()[0]           --Read a byte to determine the packet type.
    local t = packet[ netname ]
    if not t then
      return nil, "Malformed packet. Unknown header:\n"..readBuffer:get()
    end
    if #readBuffer < t.size then 
      return nil, "Malformed packet. Packet too small:\n"..readBuffer:get()
    end   --Malformed packets might cause an overread.
    
    --Allocate new struct and copy into it.
    n = n + 1
    data[n] = ffi.new( t.ct )
    types[n] = t.name
    ffi.copy( data[n], readBuffer:ref(), t.size )
    readBuffer:skip( t.size )
  end

  return data, types
end

function packet.getString( member )
  return ffi.string( member, ffi.sizeof( member ) )
end

--Slow!
function packet.luaString( member )
  local s = packet.getString( member )
  return s:sub( 1, ( s:find( "\0" ) or s:len() ) - 1 )
end

local writeBuffer = buffer.new( 1024 )
function packet.add( struct, data )
  local str = ffi.new( struct.ct, data or 0 )
  str.netname = assert( struct.netname )
  writeBuffer:putcdata( str, struct.size )
  return str
end

function packet.get()
  return writeBuffer:get()
end

mt.__call = packet.add

local testing = testing
--TESTS--
if testing then

  packet.serverInfo{
    players = 0,
    capacity = 255,
    map = "abcdefghijklmnopqrstuvwxyz1234567890",
    svname = "😘😘😘😘😘😘😘kissyfaceserver",
    version = 25,
    port = 51312,
    ip = ipString.new '132.145.25.62'
  }

  packet.heartbeat{
    tick = 49,
    hash = 33753745832876,
    protocol = 25
  }

  packet.insect{
    id = 5,
    dead = true,
    hp = -3,
    vx = -5,
    vy = 47,
    x = 59183,
    y = 21412
  }
  
  local d, t = packet.deserialise( packet.get() )
  assert( #writeBuffer == 0, "Test failed. Write buffer not empty!" )
  for i = 1, #d do
    print( "", t[i], d[i] )
    if t[i] == 'serverInfo' then
      print( "", "", packet.getString( d[i].map ), packet.getString( d[i].svname ) )
    end
    if t[i] == 'insect' then print( d[i].vx ) end
  end
  
  packet[42]{}
  
  assert( not( pcall( packet.deserialise, "grgrsgs" ) ), "Test failed. Failed to reject malformed packet." )
end
--END TESTS--

return packet